package dev.fitko.fitconnect.core;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventLog;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventStatus;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationContext;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.*;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.destination.DestinationService;
import dev.fitko.fitconnect.api.services.destination.KeyService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.events.EventLogVerificationService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.core.util.EventLogUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

public class SubmissionSender implements Sender {

    private static final Logger LOGGER = LoggerFactory.getLogger(SubmissionSender.class);

    private static final ObjectMapper MAPPER = new ObjectMapper();

    private final ValidationService validationService;
    private final CryptoService cryptoService;
    private final SubmissionService submissionService;
    private final EventLogService eventLogService;
    private final EventLogVerificationService eventLogVerifier;
    private final KeyService keyService;
    private final DestinationService destinationService;

    public SubmissionSender(final DestinationService destinationService,
                            final SubmissionService submissionService,
                            final EventLogService eventLogService,
                            final EventLogVerificationService eventLogVerifier,
                            final CryptoService encryptionService,
                            final ValidationService validationService,
                            final KeyService keyService) {

        this.submissionService = submissionService;
        this.destinationService = destinationService;
        this.eventLogService = eventLogService;
        this.eventLogVerifier = eventLogVerifier;
        this.cryptoService = encryptionService;
        this.validationService = validationService;
        this.keyService = keyService;
    }

    @Override
    public ValidationResult validatePublicKey(final RSAKey publicKey) {
        LOGGER.info("Validating public key integrity");
        return this.validationService.validateEncryptionPublicKey(publicKey);
    }

    @Override
    public ValidationResult validateMetadata(final Metadata metadata) {
        return this.validationService.validateMetadataSchema(metadata);
    }

    @Override
    public String encryptBytes(final RSAKey publicKey, final byte[] data) {
        LOGGER.info("Encrypting {} bytes", data.length);
        return this.cryptoService.encryptBytes(publicKey, data);
    }

    @Override
    public String createHash(final byte[] data) {
        LOGGER.info("Creating hash of {} bytes", data.length);
        return this.cryptoService.hashBytes(data);
    }

    @Override
    public SubmissionForPickup createSubmission(final CreateSubmission submission) {
        LOGGER.info("Creating new submission for destination {}", submission.getDestinationId());
        return this.submissionService.announceSubmission(submission);
    }

    @Override
    public void uploadAttachment(final UUID submissionId, final UUID attachmentId, final String encryptedAttachment) {
        LOGGER.info("Uploading attachment {} for submission {}", attachmentId, submissionId);
        this.submissionService.uploadAttachment(submissionId, attachmentId, encryptedAttachment);
    }

    @Override
    public Destination getDestination(final UUID destinationId) {
        return this.destinationService.getDestination(destinationId);
    }

    @Override
    public RSAKey getEncryptionKeyForDestination(final UUID destinationId) {
        LOGGER.info("Getting encryption key for destination id {}", destinationId);
        try {
            return this.keyService.getPublicEncryptionKey(destinationId);
        } catch (final RestApiException | InvalidKeyException e) {
            throw new KeyNotRetrievedException("Retrieving encryption JWK failed", e);
        }
    }

    @Override
    public String encryptObject(final RSAKey encryptionKey, final Object obj) {
        try {
            return this.cryptoService.encryptString(encryptionKey, MAPPER.writeValueAsString(obj));
        } catch (final JsonProcessingException | EncryptionException e) {
            throw new EncryptionException("Failure encrypting object", e);
        }
    }

    @Override
    public Submission sendSubmission(final SubmitSubmission submission) {
        LOGGER.info("Sending submission {}", submission.getSubmissionId());
        return this.submissionService.sendSubmission(submission);
    }

    @Override
    public ValidationResult validateJsonFormat(final String json) {
        return this.validationService.validateJsonFormat(json);
    }

    @Override
    public ValidationResult validateXmlFormat(final String xml) {
        return this.validationService.validateXmlFormat(xml);
    }

    @Override
    public List<EventLogEntry> getEventLog(final UUID caseId, final UUID destinationId) {
        final EventLog eventLog = this.eventLogService.getEventLog(caseId);
        final List<SignedJWT> events = EventLogUtil.getJWTSFromEvents(eventLog.getEventLogs());
        final ValidationContext ctx = new ValidationContext(destinationId, caseId);
        validateLog(this.eventLogVerifier.validateEventLogs(ctx, events));
        return EventLogUtil.mapEventLogToEntries(eventLog);
    }

    @Override
    public EventStatus getLastedEvent(final UUID destinationId, final UUID caseId, final UUID submissionId, final AuthenticationTags authenticationTags) {
        final EventLog eventLog = this.eventLogService.getEventLog(caseId);
        final SignedJWT latestEvent = getLatestEventForSubmission(submissionId, eventLog);
        final var ctx = new ValidationContext(destinationId, caseId, authenticationTags);
        validateLog(this.eventLogVerifier.validateEventLogs(ctx, List.of(latestEvent)));
        final EventLogEntry eventLogEntry = EventLogUtil.eventToEntry(latestEvent);
        return new EventStatus(eventLogEntry.getEvent().getState(), eventLogEntry.getProblems());
    }

    private static SignedJWT getLatestEventForSubmission(final UUID submissionId, final EventLog eventLog) {
        final SortedMap<Event, SignedJWT> logsByEvent = new TreeMap<>(EventLogUtil.getJWTSFromEvents(eventLog.getEventLogs())
                .stream()
                .filter(EventLogUtil.eventEqualsSubmissionId(submissionId))
                .collect(Collectors.toMap(EventLogUtil::getEventFromJWT, Function.identity())));

        final Iterator<Event> iter = logsByEvent.keySet().iterator();
        Event current = iter.next();
        while (iter.hasNext()) {
            final Event next = iter.next();
            if (!current.allowedNextEvents().contains(next)) {
                throw new EventLogException("Event state transition " + current.getState() + " => " + next.getState() + " is not allowed");
            }
            current = next;
        }
        return logsByEvent.get(current);
    }

    private static void validateLog(final List<ValidationResult> results) {
        if (!results.isEmpty()) {
            final String errorMessages = results.stream()
                    .map(ValidationResult::getError)
                    .map(Throwable::getMessage)
                    .distinct()
                    .collect(Collectors.joining("\n"));
            throw new EventLogException("Event log contains invalid entries: \n" + errorMessages);
        }
    }
}
