package dev.fitko.fitconnect.core;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.domain.model.event.EventLog;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.AttachmentsMismatch;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionsForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.DataIntegrityException;
import dev.fitko.fitconnect.api.exceptions.EventLogException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.ValidationException;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.events.EventLogVerificationService;
import dev.fitko.fitconnect.api.services.events.SecurityEventService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.text.ParseException;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyInt;
import static org.mockito.Mockito.*;

class SubmissionSubscriberTest {

    private SubmissionService submissionServiceMock;
    private EventLogService eventLogServiceMock;

    private EventLogVerificationService eventLogVerifierMock;
    private CryptoService cryptoServiceMock;
    private ValidationService validationServiceMock;
    private SecurityEventService securityEventServiceMock;

    private Subscriber underTest;

    @BeforeEach
    public void setUp() {

        this.submissionServiceMock = mock(SubmissionService.class);
        this.eventLogServiceMock = mock(EventLogService.class);
        this.eventLogVerifierMock = mock(EventLogVerificationService.class);
        this.cryptoServiceMock = mock(CryptoService.class);
        this.validationServiceMock = mock(ValidationService.class);
        this.securityEventServiceMock = mock(SecurityEventService.class);

        this.underTest = new SubmissionSubscriber(
                this.submissionServiceMock,
                this.eventLogServiceMock,
                this.eventLogVerifierMock,
                this.cryptoServiceMock,
                this.validationServiceMock,
                this.securityEventServiceMock
        );
    }

    @Test
    void decryptStringContent() throws JOSEException {

        // Given
        final RSAKey privateKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate();

        final var expectedDecryptedString = "decryptedString";

        when(this.cryptoServiceMock.decryptBytes(any(), any())).thenReturn(expectedDecryptedString.getBytes());

        // When
        final byte[] decryptedBytes = this.underTest.decryptStringContent(privateKey, "encryptedString");

        // Then
        assertThat(new String(decryptedBytes), equalTo(expectedDecryptedString));
    }

    @Test
    void pollAvailableSubmissions() {

        // Given
        final var destinationId = UUID.randomUUID();

        final SubmissionForPickup submission = new SubmissionForPickup();
        final SubmissionsForPickup availableSubmissions = new SubmissionsForPickup();
        availableSubmissions.setSubmissions(Set.of(submission));

        when(this.submissionServiceMock.pollAvailableSubmissions(any(), anyInt(), anyInt())).thenReturn(availableSubmissions);

        // When
        final Set<SubmissionForPickup> submissionForPickups = this.underTest.pollAvailableSubmissions(destinationId, 0, 100);

        // Then
        assertThat(submissionForPickups, equalTo(availableSubmissions.getSubmissions()));
        assertThat(submissionForPickups.size(), is(availableSubmissions.getSubmissions().size()));
    }

    @Test
    void getSubmission() {

        // Given
        final var submissionId = UUID.randomUUID();

        final var expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(submissionId);

        when(this.submissionServiceMock.getSubmission(any())).thenReturn(expectedSubmission);

        // When
        final Submission submission = this.underTest.getSubmission(submissionId);

        // Then
        assertNotNull(submission);
        assertThat(submission.getSubmissionId(), is(submissionId));
    }

    @Test
    void submissionNotFound() {

        // Given
        when(this.submissionServiceMock.getSubmission(any())).thenThrow(RestApiException.class);

        // Then
        assertThrows(
                RestApiException.class,
                () -> this.underTest.getSubmission(UUID.randomUUID()),
                "Expected getSubmission() to throw, but nothing was thrown"
        );
    }

    @Test
    void fetchAttachment() {

        // Given
        final var submissionId = UUID.randomUUID();
        final var attachmentId = UUID.randomUUID();
        final var expectedAttachment = "encryptedAttachmentContent";

        when(this.submissionServiceMock.getAttachment(any(), any())).thenReturn(expectedAttachment);

        // When
        final String encryptedAttachment = this.underTest.fetchAttachment(submissionId, attachmentId);

        // Then
        assertThat(encryptedAttachment, equalTo(expectedAttachment));
    }

    @Test
    void validateMetadata() {

        // Given
        when(this.validationServiceMock.validateMetadataSchema(any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = this.underTest.validateMetadata(new Metadata());

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void invalidMetadata() {

        // Given
        when(this.validationServiceMock.validateMetadataSchema(any())).thenReturn(ValidationResult.error(new Exception("Failed parsing metadata")));

        // When
        final ValidationResult validationResult = this.underTest.validateMetadata(new Metadata());

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Failed parsing metadata"));
    }

    @Test
    void validateHashIntegrity() {

        when(this.validationServiceMock.validateHashIntegrity(any(), any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = this.underTest.validateHashIntegrity("originalHex", "data".getBytes());

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void validateCorruptHash() {

        // Given
        when(this.validationServiceMock.validateHashIntegrity(any(), any())).thenReturn(ValidationResult.error(
                new DataIntegrityException("Hash sum of transmitted data does not equal the hash of the sender.")));

        // When
        final ValidationResult validationResult = this.underTest.validateHashIntegrity("originalHexHash", "corruptData".getBytes());

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("data does not equal the hash of the sender."));
    }

    @Test
    void testAcceptSubmission() throws ParseException, IOException {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());

        final EventPayload eventPayload = new EventPayload(submission);

        final var eventJWT = getResourceAsString("/signed_jwt.txt");

        when(this.securityEventServiceMock.createAcceptSubmissionEvent(any())).thenReturn(SignedJWT.parse(eventJWT));
        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        doNothing().when(this.eventLogServiceMock).sendEvent(any(), any());

        // When
        this.underTest.acceptSubmission(eventPayload);

        // Then
        verify(this.eventLogServiceMock, times(1)).sendEvent(submission.getCaseId(), eventJWT);
    }

    @Test
    void testRejectSubmission() throws ParseException, IOException {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());

         final var problem = new AttachmentsMismatch();

        final EventPayload eventPayload = new EventPayload(submission, List.of(problem));

        final var eventJWT = getResourceAsString("/signed_jwt.txt");
        when(this.securityEventServiceMock.createRejectSubmissionEvent(any())).thenReturn(SignedJWT.parse(eventJWT));
        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        doNothing().when(this.eventLogServiceMock).sendEvent(any(), any());

        // When
        this.underTest.rejectSubmission(eventPayload);

        // Then
        verify(this.eventLogServiceMock, times(1)).sendEvent(submission.getCaseId(), eventJWT);
    }

    @Test
    void testReceiveEventLogWithValidValidation() {

        // Given
        final EventLog emptyLog = new EventLog();
        final var caseId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();

        when(this.eventLogServiceMock.getEventLog(any())).thenReturn(emptyLog);
        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());

        // When
        final List<EventLogEntry> eventLog = this.underTest.getEventLog(destinationId, caseId);

        // Then
        assertTrue(eventLog.isEmpty());
    }

    @Test
    void testReceiveEventLogWithInvalidValidation() {

        // Given
        final EventLog emptyLog = new EventLog();
        final var caseId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final ValidationResult error = ValidationResult.error(new ValidationException("Invalid log"));

        when(this.eventLogServiceMock.getEventLog(any())).thenReturn(emptyLog);
        when(this.eventLogVerifierMock.validateEventLogs(any(), any())).thenReturn(List.of(error));

        // Then
        assertThrows(EventLogException.class, () ->  this.underTest.getEventLog(destinationId, caseId));
    }


    private String getResourceAsString(final String filename) throws IOException {
        return new String(SubmissionSenderTest.class.getResourceAsStream(filename).readAllBytes());
    }
}